// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// package textutil contains text processing utilities.
//
// This package came to life as a result of refactoring code common to
// internal packages we have factored out of the Go repo.
package textutil
